/*GLfloat gamma = 2.0;

 *  PLY.cpp
 *  PointProcessing
 *
 *  Created by Renato Pajarola on Wed Nov 05 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 */
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <vector>

#ifdef __APPLE__
#include <glut/glut.h>
#else
#include <GL/glut.h>
#endif

#include "PLY.h"
#include "geometry.h"
using namespace std;

extern int wire;
extern int flat;
extern int light;
extern int rotating;
extern int movingLighT;

extern Vector3f light_pos, viewer_pos;
extern Vector4f light_color, ambient_light;

GLfloat diffuse[4] = {0.95, 0.95, 0.95, 1.0};
GLfloat diffuse2[4] = {0.95, 0.95, 0.95, 1.0};
GLfloat specular[4] = {0.5, 0.5, 0.5, 1.0};
GLfloat shininess[1] = {1.0};
GLfloat shininess2[1] = {1.0};
static const int imgHeight = 1024;
static const int imgWidth = 1024;
static GLubyte noise[imgHeight][imgWidth][4];
static GLubyte contrast[imgHeight][imgWidth][4];
static GLubyte contrast2[imgHeight][imgWidth][4];
static GLuint contrastTex;
static GLuint contrastTex2;
static GLuint noiseTex;
GLfloat gamma = 2.0;
float finalGray;

PLYObject::PLYObject(FILE *in)
{

	int i;
  
  nproperties = 0;
  hasnormal = hastexture = false;
  hascolor = true;

  nv = nf = 0;

  vertices = NULL;
  normals = NULL;
  colors = NULL;
  texcoords = NULL;
  faces = NULL;

  // init bounding box
  for (i = 0; i < 3; i++) {
    min[i] = FLT_MAX;
    max[i] = -FLT_MAX;
  }
  
  // default order
  for (i = 0; i < 11; i++)
    order[i] = -1;

  if (!checkHeader(in)) {
    fprintf(stderr, "Error: could not read PLY file.\n");
    return;
  }

  vertices = (Vector3f*)calloc(nv, sizeof(Vector3f));
  normals = (Vector3f*)calloc(nv, sizeof(Vector3f));
  colors = (Color3u*)calloc(nv, sizeof(Color3u));
  if (hastexture)
    texcoords = (Texture2f*)calloc(nv, sizeof(Texture2f));

  faces = (Index3i*)calloc(nf, sizeof(Index3i));
  fnormals = (Vector3f*)calloc(nf, sizeof(Vector3f));

  readVertices(in);
  readFaces(in);
  for (int i=0 ; i<nv ; i++)
  {
	  colors[i][0] = rand() % 256;
	  colors[i][1] = rand() % 256;
	  colors[i][2] = rand() % 256;
  }

  // $$$ CAUTION! This might flip also files which are correct
  invertNormals();
  generateTextures();
  initTextures();
}

/* This generates the charcoal texture to be used 
 * by the model. This is procedurally generated according to 
 * contrast enhancement exponent (gamma). The texture is darker and
 * has more grains at the bottom, and is lighter with less grains
 * at the top.
 */
void PLYObject::generateTextures()
{	
	for(int i = 0; i < imgHeight; i++)
	{
		float ceo = (float) (i+1) / (float) imgHeight;
		for(int j = 0; j < imgWidth; j++)
		{
			int random = rand() % 100;
			double color = 0.0; 
			if(random < 100 * ((double) i / (3*(double) imgHeight)))
				color = 1.0;

			noise[i][j][0] = (GLubyte) color*255;
			noise[i][j][1] = (GLubyte) color*255;
			noise[i][j][2] = (GLubyte) color*255;
			noise[i][j][3] = (GLubyte) 255;

			double newCeo = pow(ceo, gamma);
			color += newCeo;
			if(color > 1.0)
				color = 1.0;

			double finalCol = color*255.0;
			contrast[i][j][0] = (GLubyte) finalCol;
			contrast[i][j][1] = (GLubyte) finalCol;
			contrast[i][j][2] = (GLubyte) finalCol;
			contrast[i][j][3] = (GLubyte) 255;
		}
	}

	for(int k = 0; k < 3; k++)
	{
		for(int i = 0; i < imgHeight; i++)
		{
			for(int j = 0; j < imgWidth; j++)
			{
				double avg = 0;
				int neighbors = 0;
				if(i != 0)
				{
					avg += contrast[i-1][j][0];
					neighbors++;
				}
				if(i != imgHeight-1)
				{
					avg += contrast[i+1][j][0];
					neighbors++;
				}

				if(j != 0)
				{
					avg += contrast[i][j-1][0];
					neighbors++;
				}

				if(j != imgWidth - 1)
				{
					avg += contrast[i][j+1][0];
					neighbors++;
				}

				avg = avg / neighbors;
				contrast2[i][j][0] = (GLubyte) avg;
				contrast2[i][j][1] = (GLubyte) avg;
				contrast2[i][j][2] = (GLubyte) avg;
				contrast2[i][j][3] = (GLubyte) 255;
			}
		}

		for(int i = 0; i < imgHeight; i++)
			for(int j = 0; j < imgWidth; j++)
			{
				contrast[i][j][0] = contrast2[i][j][0];
				contrast[i][j][1] = contrast2[i][j][1];
				contrast[i][j][2] = contrast2[i][j][2];
				contrast[i][j][3] = contrast2[i][j][3];
			}

	}
}

/*
 * Binds the data generated in generateTextures() to a texture.
 */ 
void PLYObject::initTextures()
{
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &contrastTex);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, contrastTex);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 
				   GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
				   GL_NEAREST);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, imgWidth, 
				imgHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 
				contrast2);
}

PLYObject::~PLYObject()
{
  // delete all allocated arrays

  if (vertices)
    free(vertices);
  if (normals)
    free(normals);
  if (colors)
    free(colors);
  if (texcoords)
    free(texcoords);
  if (faces)
    free(faces);  
}


bool PLYObject::checkHeader(FILE *in)
{
  char buf[128], type[128], c[32];
  int i;

  // read ply file header
  fscanf(in, "%s\n", buf);
  if (strcmp(buf, "ply") != 0) {
    fprintf(stderr, "Error: Input file is not of .ply type.\n");
    return false;
  }
  fgets(buf, 128, in);
  if (strncmp(buf, "format ascii", 12) != 0) {
    fprintf(stderr, "Error: Input file is not in ASCII format.\n");
    return false;
  }

  fgets(buf, 128, in);
  while (strncmp(buf, "comment", 7) == 0)
    fgets(buf, 128, in);

  // read number of vertices
  if (strncmp(buf, "element vertex", 14) == 0)
    sscanf(buf, "element vertex %d\n", &nv);
  else {
    fprintf(stderr, "Error: number of vertices expected.\n");
    return false;
  }

  // read vertex properties order
  i = 0;
  fgets(buf, 128, in);  
  while (strncmp(buf, "property", 8) == 0) {
    sscanf(buf, "property %s %s\n", type, c);
    if (strncmp(c, "x", 1) == 0)
      order[0] = i;
    else if (strncmp(c, "y", 1) == 0)
      order[1] = i;
    else if (strncmp(c, "z", 1) == 0)
      order[2] = i;

    else if (strncmp(c, "nx", 2) == 0)
      order[3] = i;
    else if (strncmp(c, "ny", 2) == 0)
      order[4] = i;
    else if (strncmp(c, "nz", 2) == 0)
      order[5] = i;

    else if (strncmp(c, "red", 3) == 0)
      order[6] = i;
    else if (strncmp(c, "green", 5) == 0)
      order[7] = i;
    else if (strncmp(c, "blue", 4) == 0)
      order[8] = i;

    else if (strncmp(c, "tu", 2) == 0)
      order[9] = i;
    else if (strncmp(c, "tv", 2) == 0)
      order[10] = i;

    i++;
    fgets(buf, 128, in);
  }
  nproperties = i;

  for (i = 0; i < 3; i++) {
    if (order[i] < 0) {
      fprintf(stderr, "Error: not enough vertex coordinate fields (nx, ny, nz).\n");
      return false;
   }
  }
  hasnormal = true;
  for (i = 3; i < 6; i++)
    if (order[i] < 0)
      hasnormal = false;
  hascolor = true;
  for (i = 6; i < 9; i++)
    if (order[i] < 0)
      hascolor = false;
  hastexture = true;
  for (i = 9; i < 11; i++)
    if (order[i] < 0)
      hastexture = false;

  if (!hasnormal)
    fprintf(stderr, "Warning: no normal coordinates used from file.\n");
  if (!hascolor)
    fprintf(stderr, "Warning: no color used from file.\n");
  if (!hastexture)
    fprintf(stderr, "Warning: no texture coordinates used from file.\n");

  // number of faces and face properties
  if (strncmp(buf, "element face", 12) == 0)
    sscanf(buf, "element face %d\n", &nf);
  else {
    fprintf(stderr, "Error: number of faces expected.\n");
    return false;
  }

  fgets(buf, 128, in);
  if (strncmp(buf, "property list", 13) != 0) {
    fprintf(stderr, "Error: property list expected.\n");
    return false;
  }

  fgets(buf, 128, in);
  while (strncmp(buf, "end_header", 10) != 0)
    fgets(buf, 128, in);

  return true;
}


void PLYObject::readVertices(FILE *in)
{
  char buf[128];
  int i, j;
  float values[32];
  
  // read in vertex attributes
  for (i = 0; i < nv; i++) {
    fgets(buf, 128, in);
    sscanf(buf,"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", &values[0], &values[1], &values[2], &values[3],
            &values[4], &values[5], &values[6], &values[7], &values[8], &values[9], &values[10], &values[11],
           &values[12], &values[13], &values[14], &values[15]);

    for (j = 0; j < 3; j++)
      vertices[i][j] = values[order[j]];
    if (hasnormal)
      for (j = 0; j < 3; j++)
        normals[i][j] = values[order[3+j]];
    if (hascolor)
      for (j = 0; j < 3; j++)
        colors[i][j] = (unsigned char)values[order[6+j]];
    if (hastexture)
      for (j = 0; j < 2; j++)
        texcoords[i][j] = values[order[9+j]];

    for (j = 0; j < 3; j++) {
      if (vertices[i][j] < min[j])
        min[j] = vertices[i][j];
      if (vertices[i][j] > max[j])
        max[j] = vertices[i][j];
    }
  }
}


void PLYObject::readFaces(FILE *in)
{
  char buf[128];
  int i, j, k;
  
  // read in face connectivity
  for (i = 0; i < nf; i++) {
    fgets(buf, 128, in);
    sscanf(buf, "%d %d %d %d", &k, &faces[i][0], &faces[i][1], &faces[i][2]);
    if (k != 3) {
      fprintf(stderr, "Error: not a triangular face.\n");
      exit(1);
    }
    // set up face normal
    normal(fnormals[i], vertices[faces[i][0]], vertices[faces[i][1]], vertices[faces[i][2]]);

    // accumulate normal information of each vertex
    if (!hasnormal)
      for (j = 0; j < 3; j++)
        for (k = 0; k < 3; k++)
          normals[faces[i][j]][k] += fnormals[i][k];
  }
  if (!hasnormal)
    for (i = 0; i < nv; i++)
      normalize(normals[i]);
  hasnormal = true;
}


void PLYObject::resize()
{
  int i;
  float minx, miny, minz, maxx, maxy, maxz;
  float size, scale;

  // get bounding box
  minx = miny = minz = FLT_MAX;
  maxx = maxy = maxz = -FLT_MAX;
  for (i = 0; i <nv; i++) {
    if (vertices[i][0] < minx)
      minx = vertices[i][0];
    if (vertices[i][0] > maxx)
      maxx = vertices[i][0];

    if (vertices[i][1] < miny)
      miny = vertices[i][1];
    if (vertices[i][1] > maxy)
      maxy = vertices[i][1];

    if (vertices[i][2] < minz)
      minz = vertices[i][2];
    if (vertices[i][2] > maxz)
      maxz = vertices[i][2];
  }

  // rescale vertex coordinates to be in [-1,1]
  size = 0.0;
  if (size < maxx-minx)
    size = maxx-minx;
  if (size < maxy-miny)
    size = maxy-miny;
  if (size < maxz-minz)
    size = maxz-minz;
  scale = 2.0 / size;
  for (i = 0; i < nv; i++) {
    vertices[i][0] = scale * (vertices[i][0] - minx) - 1.0;
    vertices[i][1] = scale * (vertices[i][1] - miny) - 1.0;
    vertices[i][2] = scale * (vertices[i][2] - minz) - 1.0;
  }
}

void PLYObject::invertNormals()
{
  int i, tmp;

  for (i = 0; i < nv; i++)
    scale(normals[i], -1.0);
  for (i = 0; i < nf; i++) {
    scale(fnormals[i], -1.0);
    tmp = faces[i][0];
    faces[i][0] = faces[i][2];
    faces[i][2] = tmp;
  }
}


/**************  TRYME-3 **************/
/* Make the appropriate change in this function to render the wireframe 
   in Blue instead of black
	 HINT: Colors in OpenGL are specified with 3 numbers.
	       (RGB format, for red, green and blue.)
 */

void PLYObject::draw(int index)
{
  // setup default material
  if(!index)
  {
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
	  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
	  glColor3fv(diffuse);
  }
  else
  {
  	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse2);
	  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess2);
	  glColor3fv(diffuse2);
  }

  // setup per-point color mode
  if (light)
    {
      glDisable(GL_COLOR_MATERIAL);
    }
  else
    {
      glEnable(GL_COLOR_MATERIAL);
	  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    }

  // render wireframe model
  if (wire) 
    {
      glDisable(GL_LIGHTING);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glColor3f(0.0, 0.0, 1.0);
      
      glBegin(GL_TRIANGLES);
      for (int i = 0; i < nf; i++)
        for (int j = 0; j < 3; j++)
          glVertex3fv(vertices[faces[i][j]]);	// vertex coordinates
      glEnd();
      
      glEnable(GL_POLYGON_OFFSET_FILL);
      glPolygonOffset(0.5, 1.0);
    }
  
  // set lighting if enabled
  if (light)
    glEnable(GL_LIGHTING);
  else
    glDisable(GL_LIGHTING);
  
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glBindTexture(GL_TEXTURE_2D, contrastTex);
  glBegin(GL_TRIANGLES);
  
  for (int i = 0; i < nf; i++) 
    {
      for (int j = 0; j < 3; j++) 
        {
          if (light) 
			{
			  /*
			   * render using OpenGL lighting
			   */
			   
			   glNormal3fv(normals[faces[i][j]]);	// vertex normal
			} 
		  else
		  {
			  // $$$ Compute vertex lighting here, using its normal (normals[faces[i][j]])
			  // ...
			  // Note that you have to scale the [0,1] range to [0,255] because colors has unsigned char type.

			  /*
			   * There is leftover code here from an earlier project that calculated color.
			   * This applies the same contrast enhancement to the model, and uses the color value to index the uv
			   * texture positons. The texture is then modulated with the model color.
			   */
				Vector3f h;
				Vector3f diff;
				Vector3f spec;
				Vector3f final;
				Vector3f amb;
				Vector3f lp = {light_pos[0]-vertices[faces[i][j]][0], light_pos[1]-vertices[faces[i][j]][1], light_pos[2]-vertices[faces[i][j]][2]};
				Vector3f vp = {viewer_pos[0]-vertices[faces[i][j]][0], viewer_pos[1]-vertices[faces[i][j]][1], viewer_pos[2]-vertices[faces[i][j]][2]};
				Vector3f norm = {normals[faces[i][j]][0], normals[faces[i][j]][1], normals[faces[i][j]][2]};


				normalize(lp);
				normalize(vp);
				add(h, lp, vp);
				normalize(h);

				float dpd = dotProd(lp, norm);
				float dps = dotProd(h, norm);
				if(dpd>1)
					dpd=1;
				if(dpd<0)
					dpd=0;
				if(dps>1)
					dps=1;
				if(dps<0)
					dps=0;

				double power;
				if(!index)
					 power = pow(dps, shininess[0]);
				else
					 power = pow(dps, shininess2[0]);

				if(!index)
				{
					for(int m=0; m<3; m++)
					{
						amb[m] = ambient_light[m] * diffuse[m];
						diff[m] = dpd * light_color[m] * diffuse[m];
						spec[m] = power * light_color[m] * specular[m];
						final[m] = diff[m] + spec[m] + amb[m];
						final[m] = pow(final[m], gamma);

					}	
						
						finalGray = final[0];
						if(finalGray < 0)
							finalGray = 0;
						if(finalGray > 1)
							finalGray = 1;
					    scale(final, 255);

						for(int i=0; i<3; i++)
						{
							if(final[i]>255)
								final[i] = 255;
							if(final[i]<0)
								final[i] = 0;
						}

						GLubyte fin[3] = {final[0], final[1], final[2]};
						glColor3ubv(fin);
						
				}

				else
				{
					for(int m=0; m<3; m++)
					{
						amb[m] = ambient_light[m] * diffuse2[m];
						diff[m] = dpd * light_color[m] * diffuse2[m];
						spec[m] = power * light_color[m] * specular[m];
						final[m] = diff[m] + spec[m] + amb[m];
						final[m] = pow(final[m], gamma);
					}
					finalGray = final[0];
					if(finalGray < 0)
						finalGray = 0;
					if(finalGray > 1)
						finalGray = 1;
					scale(final, 255);

					for(int i=0; i<3; i++)
						{
							if(final[i]>255)
								final[i] = 255;
							if(final[i]<0)
								final[i] = 0;
						}
					GLubyte fin[3] = {final[0], final[1], final[2]};
					glColor3ubv(fin);				
				}
		  }
		  glTexCoord2f(0.5*j, finalGray);	  
          glVertex3fv(vertices[faces[i][j]]);	// vertex coordinates
        }
    }
  glEnd();
  //glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glDisable(GL_POLYGON_OFFSET_FILL);
  if (hascolor)
    glDisable(GL_COLOR_MATERIAL);
}


void PLYObject::eat()
{
  float scale = 0.01;

  /* Jitter each vertex */
  for(int i=0 ; i<nv ;i++)
    for(int j=0 ; j<3 ; j++)
      {
        vertices[i][j] += scale*normals[i][j];
      }
}


void PLYObject::starve()
{
  float scale = -0.01;

  /* Jitter each vertex */
  for(int i=0 ; i<nv ;i++)
    for(int j=0 ; j<3 ; j++)
      {
        vertices[i][j] += scale*normals[i][j];
      }
}


/**************  TRYME-4 **************/
/* Translate all vertices in the PLYObject by a random vector  
   and watch your bunny dance */


void PLYObject::dance()
{
  /* This creates a random vector */
  Vector3f randomvector;
  randomvector[0] = rangerand(-0.1, 0.1, 30);
  randomvector[1] = rangerand(-0.1, 0.1, 30);
  randomvector[2] = rangerand(-0.1, 0.1, 30);


  /* Insert code here to add randomvector to all vertices in the PLYObject */
  /* Iterate over all the vertices and add randomvector to each one. */
  /* The vertex positions are stored in the array called 'vertices' */
  for(int i=0 ; i<nv ;i++)
    for(int j=0 ; j<3 ; j++)
		vertices[i][j] += randomvector[j];
}



/* A not too good pseudo-random number generator */
/* Returns a pseudo-random number between min and max */
double PLYObject::rangerand(double min, double max, long steps)
{
  return min + ((rand() % steps) * (max - min)) / steps;
}