#include <signal.h>
#include <sys/types.h>
#include <math.h>
#include "geometry.h"
#include "inputModule.h"
#include "PLY.h"

/* This File contains the KeyBoard and mouse handling routines */


static int motionMode;
static int startX;
static int startY;
static GLfloat angle = 20;    /* in degrees */
static GLfloat angle2 = 30;   /* in degrees */
GLfloat lightAngle = 20;
GLfloat lightAngle2 = 30;

GLfloat current_pos[] = {0.0, 0.0, 5.0};

int wire = 0;
int flat = 0;
int light = 0;
int rotating = 1;
int movingLighT = 0;

extern PLYObject* ply[2];
extern Vector3f light_pos, viewer_pos;

void readKeyboard(unsigned char key, int x, int y)
{
  switch(key){
  case  0x1B:
  case  'q':
  case  'Q':
    exit(0);
    break; 
  case '+':
    if (ply)
	{
      ply[0]->eat();
	  ply[1]->eat();
	}
    break;
  case '-':
    if (ply)
	{
      ply[0]->starve();
	  ply[1]->starve();
	}
    break;
  case 'd':
    if (ply)
	{
      ply[0]->dance();
	  ply[1]->dance();
	}
    break;
  case 'w':
  case 'W':
    wire = (wire + 1) % 2;
    break;
  case 'f':
  case 'F':
    flat = (flat + 1) % 2;
    break;
  case 'i':
  case 'I':
    if (ply)
	{
      ply[0]->invertNormals();
	  ply[1]->invertNormals();
	}
    break;
  case 'l':
  case 'L':
    light = (light + 1) % 2;
    break;
  case 'a':
  case 'A':
    rotating = (rotating + 1) % 2;
    break;
  case 't':
  case 'T':
    movingLighT = (movingLighT + 1) % 2;
    break;
  case 'r':
  case 'R':
    // reset initial view parameters
    angle = 20;
    angle2 = 30;
    current_pos[0] = 0.0;
    current_pos[1] = 0.0;
    current_pos[2] = 5.0;
    break;
  case 'h':
  case 'H':
    printf("\tPress q/Q for Quit\n");
    printf("\tPress h/H to print this help\n");
    printf("\tPress l/L to turn on/off Lighting\n");
    printf("\tPress w/W to toggle WireFrame Rendering\n");
    printf("\tPress i/I to invert the normals\n");
    printf("\tPress f/F to toggle Flat Shading\n");
    printf("\tPress r/R to revert ViewPoint to initial position\n");
    printf("\tPress + to make the bunny grow fatter\n");
    printf("\tPress - to make the bunny grow thinner\n");
    printf("\tPress d/D to make the bunny dance randomly\n");
	printf("\tPress t/T to move the light position\n");
	printf("\tPress arrow keys to move camera position\n");
  default:
    break;
  }
  glutPostRedisplay();
}


void readSpecialKeys(int key, int x, int y)
{
  switch(key){
  case GLUT_KEY_UP:
	current_pos[2]-= .5;
    break;
  case GLUT_KEY_DOWN: 
    current_pos[2]+= .5;
    break;
  case GLUT_KEY_RIGHT: 
	  current_pos[0]+= .5;
    break;
  case GLUT_KEY_LEFT:  
		current_pos[0]-= .5;
    break;
  }
  glutPostRedisplay();
}


void mouseButtHandler(int button, int state, int x, int y)
{
  motionMode = 0;

  switch(button){
  case GLUT_LEFT_BUTTON:  
    if(state == GLUT_DOWN) {
      motionMode = 1;		// Rotate object
      startX = x;
      startY = y;
    }
    break;
  case GLUT_MIDDLE_BUTTON:  
    if(state == GLUT_DOWN) {
      motionMode = 2;		// Translate object
      startX = x;
      startY = y;
    }
    break;
  case GLUT_RIGHT_BUTTON: 
    if(state == GLUT_DOWN) { 
      motionMode = 3;		// Zoom
      startX = x;
      startY = y;
    }
    break;
  }
  glutPostRedisplay();
}


void mouseMoveHandler(int x, int y)
{
  
  // No mouse button is pressed... return 
  switch(motionMode){
  case 0:
	  return;
    break;

  case 1: // Calculate the rotations
	  if(movingLighT)
	  {
		  lightAngle = lightAngle + (x - startX);
		  lightAngle2 = lightAngle2 + (y - startY);
		double sineX = sin(lightAngle2 * M_PI/45);
		double sineY = sin(lightAngle * M_PI/45);
		double cosineX = cos(lightAngle2);
		double cosineY = cos(lightAngle);

		Matrix4f rotX = {cosineX, -sineX, 0, 0, 
						sineX, cosineX, 0, 0, 
						0, 0, 1, 0, 
						0, 0, 0, 1};
		Matrix4f rotY = {cosineY, 0, -sineY, 0,
						0, 1, 0, 0,
						sineY, 0, cosineY, 0,
						0, 0, 0, 1};
		multVector(light_pos, rotX, light_pos);
		multVector(light_pos, rotY, light_pos);
		glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	  }  
	  else
	  {
		angle = angle + (x - startX);
		angle2 = angle2 + (y - startY);
	  }
    startX = x;
    startY = y;
    break;

  case 2:
	  if(movingLighT)
	  {
		light_pos[0] = light_pos[0] - (x - startX);
		light_pos[1] = light_pos[1] - (y - startY);  
		glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	  }
	  else
	  {
		current_pos[0] = current_pos[0] - (x - startX)/100.0;
		current_pos[1] = current_pos[1] - (y - startY)/100.0;
	  }
    startX = x;
    startY = y;
    break;
    
  case 3:
	  if(movingLighT)
	  {
		light_pos[2] = light_pos[2] - (y - startY);
	    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	  }

	  else
		current_pos[2] = current_pos[2] - (y - startY)/10.0;
    startX = x;
    startY = y;
    break;
  }
  
  glutPostRedisplay();
}

void passiveMouseMoveHandler(int x, int y)
{
	startX = x;
	startY = y;
	gluLookAt(current_pos[0], current_pos[1], current_pos[2], startX - x, startY - y, 1, 0,1,0);
	glutPostRedisplay();
}

void setUserView()
{
  glLoadIdentity();

  glTranslatef(-current_pos[0], current_pos[1], -current_pos[2]);
  glRotatef(angle2, 1.0, 0.0, 0.0);
  glRotatef(angle, 0.0, 1.0, 0.0);
}

void incrementAngle(int increment)
{
	angle += increment;
	glutPostRedisplay();
}

GLfloat getMyAngle()
{
	return angle;
}

void setMyAngle(GLfloat ang)
{
	angle = ang;
}