#include "time.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#ifdef _WIN32
#include <windows.h>
#endif
#ifdef __APPLE__
#include <opengl/gl.h>
#include <glut/glut.h>
#else
#include <GL/gl.h>
#include <GL/glut.h>
#endif

#include <iostream>
#include "viewModule.h"
#include "inputModule.h"
#include "cube.h"
#include "PLY.h"
#include "geometry.h"
using namespace std;

int window;
int updateFlag;
extern int rotating;
extern int movingLighT;

void timerFunc(int value);
PLYObject *ply[2];


perspectiveData pD;

Vector4f initial_light_pos = {-100.0, 100.0, 100.0, 0.0};
Vector4f ambient_light = {0.3, 0.3, 0.3, 1.0};
Vector4f light_color = {0.6, 0.6, 0.6, 1.0};
Vector4f black_color = {0.0, 0.0, 0.0, 1.0};

Vector3f light_pos, viewer_pos;
double ang = getMyAngle();

void cleanup(int sig)
{
  // insert cleanup code here (i.e. deleting structures or so)
  //fprintf(stderr,"Cleaning up\n");
  if (ply)
    delete(ply);
  exit(0);
}

//##########################################
// OpenGL Display function

void display(void)
{
  int i, j;
  float M[16];
  Matrix4f m;
  
  glutSetWindow(window);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  setUserView();

  glGetFloatv(GL_MODELVIEW_MATRIX, M);
  for (i = 0; i < 3; i++) {
    m[i][3] = 0.0;
    m[3][i] = 0.0;
    for (j = 0; j < 3; j++) {
      m[i][j] = M[i*4+j];
    }
  }
  m[3][3] = 1.0;
  multVector(light_pos, m, initial_light_pos);
  multVector(viewer_pos, m, current_pos);
  
  /**************  TRYME-2 **************/

  /* Enable rendering of the PLY when present, or the cube otherwise. */
  if (ply)
  {
	  ply[0]->draw(0);
	  //glPushMatrix();
	  //glTranslatef(3, 0, 0);
	  //ply[1]->draw(1);
	  //glPopMatrix();
  }
  else
	  drawCube();
  
  glutSwapBuffers();
}

//##########################################
// Init display settings

void initDisplay()
{
  // Perspective projection parameters
  pD.fieldOfView = 45.0;
  pD.aspect      = (float)IMAGE_WIDTH/IMAGE_HEIGHT;
  pD.nearPlane   = 0.1;
  pD.farPlane    = 50.0;

  // setup context
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(pD.fieldOfView, pD.aspect, pD.nearPlane, pD.farPlane);

  // set basic matrix mode
  glMatrixMode(GL_MODELVIEW);

  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);

  glClearColor(0.95, 0.95, 0.85, 1.0);
  glClearIndex(0);
  glClearDepth(1);

  // setup lights
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, black_color);
  glLightfv(GL_LIGHT0, GL_POSITION, initial_light_pos);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_light);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_color);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_color);

  //glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
}

//##########################################
// Main function



int main(int argc, char **argv)
{
  
  //signal(SIGHUP, cleanup);

  glutInit(&argc, argv);


  /**************  TRYME-1 **************/

  /* Uncomment this section to accept an input filename (of a PLY file)
     from the command line using the argc and argv parameters.
     Also creates a new PLY Object from the .ply file */

  FILE *in;
  char *filename;

  if (argc < 2) {
    fprintf(stderr, "Usage: %s filename\n", argv[0]);
    fprintf(stderr, "\tfilename is of PLY triangle mesh format\n");
    exit(1);
  }

  filename = argv[1];
  if (!(in = fopen(filename, "r"))) {
    fprintf(stderr, "Cannot open input file %s.\n", filename);
    exit(1);
  }

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(IMAGE_WIDTH,IMAGE_HEIGHT);
	
  window = glutCreateWindow("Final project");

  ply[0] = new PLYObject(in);
  ply[1] = ply[0];
  ply[0]->resize();
  ply[1]->resize();

  srand(time(NULL));

  glutDisplayFunc(display);
  glutKeyboardFunc(readKeyboard);
  glutSpecialFunc(readSpecialKeys);
  glutMouseFunc(mouseButtHandler);
  glutMotionFunc(mouseMoveHandler);
  glutPassiveMotionFunc(passiveMouseMoveHandler);
  glutIdleFunc(NULL);

  initDisplay();


  glutMainLoop();

  return 0;             /* ANSI C requires main to return int. */
}